import numpy as np
import os
import ctypes
import scipy
from scipy.sparse.linalg import splu
import time
import sys
import mpi4py
from mpi4py import MPI
import superlu_dist.pdbridge as pdbridge


comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

if(rank==0):
    print('mpi4py version: ', mpi4py.__version__)
    print('MPI count:', size)


####################################################################################################
####################################################################################################
####################### create the matrix
INT64 = 1 # whether to use 64bit integer (requring superlu_dist to be compiled with 64-bit indexing)
algo3d = 0 # whether to use 2D or 3D factorizations
rng = np.random.default_rng()
n = 4000
nrhs = 1
use_cov = 0

if(rank==0):
    if(use_cov==0):
        a = scipy.sparse.random(n, n, density=0.01, random_state=rng)
        m = (a.T @ a) + scipy.sparse.identity(n)
        print("sparsity: ", float(m.nnz)/n**2, "nnz(A): ", m.nnz)
    else:
        # m = scipy.sparse.load_npz('/global/cfs/cdirs/m2957/liuyangz/my_research/matrix/sparse_matrix10Mill.npz')
        # m = scipy.sparse.load_npz('/global/cfs/cdirs/m2957/liuyangz/my_research/matrix/sparse_matrix10Mill_prettydense.npz')
        m = scipy.sparse.load_npz('/global/cfs/cdirs/m2957/liuyangz/my_research/matrix/sparse_matrix10Mill_no1.npz')
        m = m.tocsr()
        m = m[0:n, 0:n]
        print("sparsity: ", float(m.nnz)/n**2, "nnz(A): ", m.nnz)
else: #dummy data, not refered inside superlu_dist API
    a = scipy.sparse.random(1, 1, density=1, random_state=rng)
    m = (a.T @ a) 
m_csc = m.tocsc()

if(INT64==0):
    rowind = m_csc.indices.astype(np.int32)
    colptr = m_csc.indptr.astype(np.int32) 
    nzval = m_csc.data.astype(np.float64)  
    nnz=m.nnz
else:
    rowind = m_csc.indices.astype(np.int64)
    colptr = m_csc.indptr.astype(np.int64) 
    nzval = m_csc.data.astype(np.float64)  
    nnz=np.int64(m.nnz)
    n = np.int64(n)




####################################################################################################
####################################################################################################
####################### handle options 
argv=sys.argv
if(len(argv)==1): # options are not passed via command line, set them manually here. If they are not set here, default values are used
    if(algo3d==1):
        argv.extend(['-d', '1'])  # process layers
        argv.extend(['-r', '%i'%(np.sqrt(size))])  # process rows
        argv.extend(['-c', '%i'%(np.sqrt(size))])  # process columns       
        argv.extend(['-b', '0'])  # batch count       
    else:
        argv.extend(['-r', '%i'%(np.sqrt(size))])  # process rows
        argv.extend(['-c', '%i'%(np.sqrt(size))])  # process columns
    
    argv.extend(['-p', '1'])  # row permutation 
    argv.extend(['-q', '2'])  # column permutation 
    argv.extend(['-s', '0'])  # parallel symbolic factorization, needs -q to be 5
    argv.extend(['-i', '0'])  # whether to use iterative refinement 0, 1, 2
    argv.extend(['-m', '0'])  # whether to use symmetric pattern 0 or 1
    argv.extend(['-n', '1'])  # whether to use tiny pivot replacement
argc = len(argv)
if(rank==0):    
    print('SuperLU options: ',argv[1:])
argv_bytes = [arg.encode('utf-8') for arg in argv]
argv_ctypes = (ctypes.c_char_p * (argc + 1))(*argv_bytes, None)






####################################################################################################
####################################################################################################
####################### call the APIs
start = time.time()
sp = pdbridge.load_library(INT64)
####################### initialization
pyobj = ctypes.c_void_p()
if(INT64==0):
    sp.pdbridge_init(
        algo3d,
        n,                              # int_t m
        n,                              # int_t n
        nnz,                            # int_t nnz
        rowind.ctypes.data_as(ctypes.POINTER(ctypes.c_int)),  # int_t *rowind
        colptr.ctypes.data_as(ctypes.POINTER(ctypes.c_int)),  # int_t *colptr
        nzval.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),  # double *nzval
        ctypes.byref(pyobj),            # void **pyobj
        argc,                           # int argc
        argv_ctypes                     # char *argv[]
    )
else:
    sp.pdbridge_init(
        algo3d,
        n,                              # int_t m
        n,                              # int_t n
        nnz,                            # int_t nnz
        rowind.ctypes.data_as(ctypes.POINTER(ctypes.c_int64)),  # int_t *rowind
        colptr.ctypes.data_as(ctypes.POINTER(ctypes.c_int64)),  # int_t *colptr
        nzval.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),  # double *nzval
        ctypes.byref(pyobj),            # void **pyobj
        argc,                           # int argc
        argv_ctypes                     # char *argv[]
    )    
end = time.time()
if(rank==0):
    print(f"Time spent in pdbridge_init: {end - start} seconds")



####################### factor 
start = time.time()
sp.pdbridge_factor(
    ctypes.byref(pyobj),            # void **pyobj
)
end = time.time()
if(rank==0):
    print(f"Time spent in pdbridge_factor: {end - start} seconds")


####################### solve 
start = time.time()
xb = np.random.rand(n*nrhs).astype(np.float64) # pdbridge_solve will broadcast xb on rank 0 to all ranks
sp.pdbridge_solve(
    ctypes.byref(pyobj),            # void **pyobj
    nrhs,                           # int nrhs
    xb.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),  # double *nzval
)
end = time.time()
if(rank==0):
    print(f"Time spent in pdbridge_solve: {end - start} seconds")



####################### log-determinant 
start = time.time()
sign = ctypes.c_int(1)
logdet = ctypes.c_double(0.0)
sp.pdbridge_logdet(
    ctypes.byref(pyobj),            # void **pyobj
    ctypes.byref(sign),                           # int nrhs
    ctypes.byref(logdet),  # double *nzval
)

if(rank==0):
    print("superlu logdet:",sign.value,logdet.value)
    # sign, logdet = np.linalg.slogdet(m.toarray())
    # print("numpy logdet:",int(sign),logdet)
end = time.time()
if(rank==0):
    print(f"Time spent in pdbridge_logdet: {end - start} seconds")


####################### free stuff
start = time.time()
sp.pdbridge_free(ctypes.byref(pyobj))
end = time.time()
if(rank==0):
    print(f"Time spent in pdbridge_free: {end - start} seconds")




import numpy as np
import scipy
from superlu_dist.pdbridge import *

######## Simulating the MCMC/LBFGS iterations 
n=1000 # starting number of samples, increment 1 each GP iteration
N_GP_iter=1
N_MCMC_iter=10
N_EI_search_iter=10

INT64 = 0 # whether to use 64bit integer (requring superlu_dist to be compiled with 64-bit indexing)
algo3d = 0 # whether to use 2D or 3D factorizations
rng = np.random.default_rng()
verbosity=True


for i in range(N_GP_iter):
    print("++ GP iteration: ", i)
    print("++ Perform MCMC ")
    for k in range(N_MCMC_iter):
        print("-- Assemble covariance matrix: ")
        a = scipy.sparse.random(n, n, density=0.01, random_state=rng)
        m = (a.T @ a) + scipy.sparse.identity(n)
        print("   sparsity: ", float(m.nnz)/n**2, "nnz(A): ", m.nnz)

        print("-- LU factor covariance matrix with ",n," samples")
        superlu_factor(m, INT64, algo3d, verbosity)

        print("-- use logdet and solve to compute likelihood: ")
        sign,logd = superlu_logdet(verbosity)
        xb = np.random.rand(n,1).astype(np.float64) 
        superlu_solve(xb, verbosity)

        # if(k != N_MCMC_iter-1):
        #     print("-- free LU factors: ") #### Don't free LU factors at the last MCMC iteration, as the LU factors are needed for performing inference once hyperparemters are optimized 
        #     superlu_freeLU(verbosity)

    print("++ Perform inference or aquization function optimization here ... ")
    for ii in range(N_EI_search_iter):
        xb = np.random.rand(n,1).astype(np.float64) 
        superlu_solve(xb, verbosity)

    print("++ Add one more sample here ... ")
    n = n + 1
    ####################################################################################################

print("++ Terminate SuperLU (after this no more superLU APIs can be called) ")
superlu_freeLU(verbosity) # this call is optional as superlu_terminate will clean up everything
superlu_terminate(verbosity)


import numpy as np
import os
import ctypes
import scipy
from scipy.sparse.linalg import splu
import time
import sys
import mpi4py
from mpi4py import MPI
import superlu_dist.pdbridge as pdbridge
import pickle


comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

if(rank==0):
    print('mpi4py version: ', mpi4py.__version__)
    print('MPI count:', size)



######################## define the files used to communicate between masters and workers
CONTROL_FILE=os.getenv("CONTROL_FILE", "control.txt")
DATA_FILE=os.getenv("DATA_FILE", "data.bin")
RESULT_FILE=os.getenv("RESULT_FILE", "result.bin")  
poll_interval = 0.001


# Ensure the file exists; if not, wait a moment and try again.
while True:
    flag=''
    if rank == 0:
        while True:
            if os.path.exists(CONTROL_FILE):
                with open(CONTROL_FILE, "r") as f:
                    flag = f.read().strip()
                if flag == "init" or flag == "factor" or flag == "solve" or flag=="logdet" or flag=="free" or flag=="terminate":
                    break
            time.sleep(poll_interval)
    flag = comm.bcast(flag, root=0)
    if(flag=="init"):
        #####  read in the matrix by rank 0
        if rank == 0:
            with open(DATA_FILE, "rb") as f:
                m,INT64,algo3d = pickle.load(f)
            n=(m.shape)[0]    
            INT64 = comm.bcast(INT64, root=0)
            algo3d = comm.bcast(algo3d, root=0)
            n = comm.bcast(n, root=0)
        else:
            INT64=-1
            algo3d=-1
            n=-1
            INT64 = comm.bcast(INT64, root=0)
            algo3d = comm.bcast(algo3d, root=0)
            n = comm.bcast(n, root=0)
            a = scipy.sparse.random(1, 1, density=1)
            m = (a.T @ a)     

        m_csc = m.tocsc()
        if(INT64==0):
            rowind = m_csc.indices.astype(np.int32)
            colptr = m_csc.indptr.astype(np.int32) 
            nzval = m_csc.data.astype(np.float64)  
            nnz=m.nnz
        else:
            rowind = m_csc.indices.astype(np.int64)
            colptr = m_csc.indptr.astype(np.int64) 
            nzval = m_csc.data.astype(np.float64)  
            nnz=np.int64(m.nnz)
            n = np.int64(n)

        ####################### handle options 
        argv=sys.argv
        if(len(argv)==1): # options are not passed via command line, set them manually here. If they are not set here, default values are used
            if(algo3d==1):
                argv.extend(['-d', '1'])  # process layers
                argv.extend(['-r', '%i'%(np.sqrt(size))])  # process rows
                argv.extend(['-c', '%i'%(np.sqrt(size))])  # process columns
                argv.extend(['-b', '0'])  # batch count             
            else:
                argv.extend(['-r', '%i'%(np.sqrt(size))])  # process rows
                argv.extend(['-c', '%i'%(np.sqrt(size))])  # process columns
            
            argv.extend(['-p', '1'])  # row permutation 
            argv.extend(['-q', '2'])  # column permutation 
            argv.extend(['-s', '0'])  # parallel symbolic factorization, needs -q to be 5
            argv.extend(['-i', '0'])  # whether to use iterative refinement 0, 1, 2
            argv.extend(['-m', '0'])  # whether to use symmetric pattern 0 or 1
            argv.extend(['-n', '1'])  # whether to use tiny pivot replacement
        argc = len(argv)
        if(rank==0):    
            print('SuperLU options: ',argv[1:])
        argv_bytes = [arg.encode('utf-8') for arg in argv]
        argv_ctypes = (ctypes.c_char_p * (argc + 1))(*argv_bytes, None)

        sp = pdbridge.load_library(INT64)
        ####################### initialization
        pyobj = ctypes.c_void_p()
        if(INT64==0):
            sp.pdbridge_init(
                algo3d,
                n,                              # int_t m
                n,                              # int_t n
                nnz,                            # int_t nnz
                rowind.ctypes.data_as(ctypes.POINTER(ctypes.c_int)),  # int_t *rowind
                colptr.ctypes.data_as(ctypes.POINTER(ctypes.c_int)),  # int_t *colptr
                nzval.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),  # double *nzval
                ctypes.byref(pyobj),            # void **pyobj
                argc,                           # int argc
                argv_ctypes                     # char *argv[]
            )
        else:
            sp.pdbridge_init(
                algo3d,
                n,                              # int_t m
                n,                              # int_t n
                nnz,                            # int_t nnz
                rowind.ctypes.data_as(ctypes.POINTER(ctypes.c_int64)),  # int_t *rowind
                colptr.ctypes.data_as(ctypes.POINTER(ctypes.c_int64)),  # int_t *colptr
                nzval.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),  # double *nzval
                ctypes.byref(pyobj),            # void **pyobj
                argc,                           # int argc
                argv_ctypes                     # char *argv[]
            )    

    elif(flag=="factor"):
        ####################### factor 
        sp.pdbridge_factor(
            ctypes.byref(pyobj),            # void **pyobj
        )
    elif(flag=="solve"):              
        ####################### solve 
        #####  read in the RHS by rank 0
        nrhs=-1  
        xb = np.random.rand(1).astype(np.float64) 
        if rank == 0:
            with open(DATA_FILE, "rb") as f:
                xb,nrhs = pickle.load(f)
            xb = np.ascontiguousarray(xb, dtype=np.float64)            

        sp.pdbridge_solve(
            ctypes.byref(pyobj),            # void **pyobj
            nrhs,                           # int nrhs
            xb.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),  # double *nzval
        )
        if rank == 0:
            with open(RESULT_FILE, "wb") as f:
                pickle.dump(xb, f)

    elif(flag=="logdet"):
        ####################### log-determinant 
        sign = ctypes.c_int(1)
        logdet = ctypes.c_double(0.0)
        sp.pdbridge_logdet(
            ctypes.byref(pyobj),            # void **pyobj
            ctypes.byref(sign),                           # int nrhs
            ctypes.byref(logdet),  # double *nzval
        )
        if rank == 0:
            with open(RESULT_FILE, "wb") as f:
                pickle.dump((sign.value, logdet.value),f)

        if(rank==0):
            print("superlu logdet:",sign.value,logdet.value)
            # sign, logdet = np.linalg.slogdet(m.toarray())
            # print("numpy logdet:",int(sign),logdet)     

    elif(flag=="free"):
        ####################### free stuff
        sp.pdbridge_free(ctypes.byref(pyobj))
    elif(flag=="terminate"):      
        break

    if(rank==0):    
        ####################### signal the master that the work (init, factor, solve, logdet, free) has been completed
        with open(CONTROL_FILE, "w") as f:
            if(flag=="free"):
                f.write("clean")                
            else:
                f.write("done")

if(rank==0):    
    ####################### signal the master that the work (terminate) has been completed and the MPI communicator will be released
    with open(CONTROL_FILE, "w") as f:
        f.write("finish")


import numpy as np
import os
import ctypes
import scipy
from scipy.sparse.linalg import splu
import time
import sys
import pickle
from superlu_dist.pdbridge import *




####################################################################################################
####################################################################################################
####################### create the matrix
INT64 = 1 # whether to use 64bit integer (requring superlu_dist to be compiled with 64-bit indexing)
algo3d = 0 # whether to use 2D or 3D factorizations
rng = np.random.default_rng()
n = 4000000
nrhs = 1
use_cov = 0
verbosity=True

if(use_cov==0):
    a = scipy.sparse.random(n, n, density=0.01, random_state=rng)
    m = (a.T @ a) + scipy.sparse.identity(n)
    print("sparsity: ", float(m.nnz)/n**2, "nnz(A): ", m.nnz)
else:
    m = scipy.sparse.load_npz('/global/cfs/cdirs/m2957/liuyangz/my_research/matrix/sparse_matrix10Mill.npz')
    # m = scipy.sparse.load_npz('/global/cfs/cdirs/m2957/liuyangz/my_research/matrix/sparse_matrix10Mill_prettydense.npz')
    # m = scipy.sparse.load_npz('/global/cfs/cdirs/m2957/liuyangz/my_research/matrix/sparse_matrix10Mill_no1.npz')
    # m = scipy.sparse.load_npz('/global/cfs/cdirs/m2957/liuyangz/my_research/matrix/sparse_matrix10Mill_no1.npz')
    m = m.tocsr()
    m = m[0:n, 0:n]
    print("sparsity: ", float(m.nnz)/n**2, "nnz(A): ", m.nnz)


####################################################################################################
####################################################################################################
####################### call the APIs

xb = np.random.rand(n,nrhs).astype(np.float64) 
superlu_factor(m, INT64, algo3d, verbosity)
sign,logd = superlu_logdet(verbosity)
superlu_solve(xb, verbosity)
superlu_freeLU(verbosity)
superlu_terminate(verbosity)



import os
import ctypes
import sys
from sys import platform
import time
import pickle
import numpy as np

def setup_pdbridge(sp, INT64):
    # Define the function signatures as shown in your original code
    sp.sizeof_int_t.argtypes = None
    sp.sizeof_int_t.restype = ctypes.c_int
    int_t = sp.sizeof_int_t()
    
    sp.pdbridge_init.restype = None
    if INT64 == 0:
        if int_t != 4:
            raise Exception("libsuperlu_dist_python has been compiled with 64-bit integers. Please recompile it with 32-bit integers or use INT64=1 in the python APIs.")

        sp.pdbridge_init.argtypes = [
            ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int,
            ctypes.POINTER(ctypes.c_int),
            ctypes.POINTER(ctypes.c_int),
            ctypes.POINTER(ctypes.c_double),
            ctypes.POINTER(ctypes.c_void_p),
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_char_p)
        ]
    else:
        if int_t != 8:
            raise Exception("libsuperlu_dist_python has been compiled with 32-bit integers. Please recompile it with 64-bit integers or use INT64=0 in the python APIs.")

        sp.pdbridge_init.argtypes = [
            ctypes.c_int, ctypes.c_int64, ctypes.c_int64, ctypes.c_int64,
            ctypes.POINTER(ctypes.c_int64),
            ctypes.POINTER(ctypes.c_int64),
            ctypes.POINTER(ctypes.c_double),
            ctypes.POINTER(ctypes.c_void_p),
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_char_p)
        ]

    sp.pdbridge_factor.restype = None
    sp.pdbridge_factor.argtypes = [ctypes.POINTER(ctypes.c_void_p)]
    sp.pdbridge_solve.restype = None
    sp.pdbridge_solve.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_int, ctypes.POINTER(ctypes.c_double)]
    sp.pdbridge_logdet.restype = None
    sp.pdbridge_logdet.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_double)]
    sp.pdbridge_free.restype = None
    sp.pdbridge_free.argtypes = [ctypes.POINTER(ctypes.c_void_p)]

def load_library(INT64):
    # Check platform and set library extension
    if platform == "linux" or platform == "linux2":
        pos = '.so'
    elif platform == "darwin":
        pos = '.dylib'
    elif platform == "win32":
        raise Exception("Windows is not yet supported")

    DLLFOUND = False
    INSTALLDIR = os.getenv('SUPERLU_PYTHON_LIB_PATH', '/usr/lib')

    DLL = os.path.abspath(__file__ + "/../../") + '/libsuperlu_dist_python' + pos
    if os.path.exists(DLL):
        DLLFOUND = True
    elif INSTALLDIR is not None:
        DLL = os.path.join(INSTALLDIR, 'libsuperlu_dist_python' + pos)
        if os.path.exists(DLL):
            DLLFOUND = True
    else:
        DLL = os.path.join('./libsuperlu_dist_python' + pos)
        if os.path.exists(DLL):
            DLLFOUND = True            
    if DLLFOUND:
        sp = ctypes.cdll.LoadLibrary(DLL)
        setup_pdbridge(sp, INT64)
        return sp
    else:
        raise Exception("Cannot find the superlu_dist_python library. Try to set the SUPERLU_PYTHON_LIB_PATH environment variable correctly.")




###################################################################################################
###########  define the APIs

def wait_for_flag(expected_flag, control_file, poll_interval=0.001):
    """Poll the control file until its content equals the expected flag."""
    while True:
        if os.path.exists(control_file):
            with open(control_file, "r") as f:
                flag = f.read().strip()
            if flag == expected_flag:
                return True
        time.sleep(poll_interval)
        

####################### initialization and factorization
def superlu_factor(KV, INT64=1, algo3d=0, verbosity=False):
    start = time.time()
    CONTROL_FILE=os.getenv("CONTROL_FILE", "control.txt")

    # The following if test makes sure superlu cleans up the factorization is there is one 
    if os.path.exists(CONTROL_FILE):
        with open(CONTROL_FILE, "r") as f:
            flag = f.read().strip()
        if flag != "clean":
            superlu_freeLU(verbosity)

    DATA_FILE=os.getenv("DATA_FILE", "data.bin")
    with open(DATA_FILE, "wb") as f:
        pickle.dump((KV,INT64,algo3d), f)
    with open(CONTROL_FILE, "w") as f:
        f.write("init")
    wait_for_flag("done", CONTROL_FILE)
    end = time.time()
    if verbosity==True:
        print(f"Time spent in pdbridge_init: {end - start} seconds")
    
    start = time.time()
    with open(CONTROL_FILE, "w") as f:
        f.write("factor")
    wait_for_flag("done", CONTROL_FILE)
    end = time.time()
    if verbosity==True:
        print(f"Time spent in pdbridge_factor: {end - start} seconds")


####################### solve 
def superlu_solve(vec, verbosity=False):
    vec = np.asarray(vec)  
    if vec.ndim == 1:
        vec = vec.reshape(-1, 1)  
    nrhs=vec.shape[-1]
    start = time.time()
    CONTROL_FILE=os.getenv("CONTROL_FILE", "control.txt")
    DATA_FILE=os.getenv("DATA_FILE", "data.bin")
    RESULT_FILE=os.getenv("RESULT_FILE", "result.bin")    
    with open(DATA_FILE, "wb") as f:
        pickle.dump((vec,nrhs), f)
    with open(CONTROL_FILE, "w") as f:
        f.write("solve")
    wait_for_flag("done", CONTROL_FILE)
    with open(RESULT_FILE, "rb") as f:
        vec_out = pickle.load(f)
        np.copyto(vec, vec_out)
    end = time.time()
    if verbosity==True:
        print(f"Time spent in pdbridge_solve: {end - start} seconds")


####################### log-determinant 
def superlu_logdet(verbosity=False):
    start = time.time()
    CONTROL_FILE=os.getenv("CONTROL_FILE", "control.txt")
    RESULT_FILE=os.getenv("RESULT_FILE", "result.bin")    
    with open(CONTROL_FILE, "w") as f:
        f.write("logdet")
    wait_for_flag("done", CONTROL_FILE)
    with open(RESULT_FILE, "rb") as f:
        sign,log_det = pickle.load(f)
    end = time.time()
    if verbosity==True:
        print(f"Time spent in pdbridge_logdet: {end - start} seconds")
    return sign,log_det

####################### free stuff
def superlu_freeLU(verbosity=False):
    start = time.time()
    CONTROL_FILE=os.getenv("CONTROL_FILE", "control.txt")
    with open(CONTROL_FILE, "w") as f:
        f.write("free")
    wait_for_flag("clean", CONTROL_FILE)
    end = time.time()
    if verbosity==True:
        print(f"Time spent in pdbridge_free: {end - start} seconds")


####################### terminate all workers if no more superLU calls are needed
def superlu_terminate(verbosity=False):
    start = time.time()
    CONTROL_FILE=os.getenv("CONTROL_FILE", "control.txt")
    with open(CONTROL_FILE, "w") as f:
        f.write("terminate")
    end = time.time()
    if verbosity==True:
        print(f"Time spent in pdbridge_terminate: {end - start} seconds")


